/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import org.w3c.dom.*;

/**
 * Attribute represents an XML-style attribute of an
 * Element. Typically, the allowable values are controlled by its
 * declaration in the Document Type Definition (DTD) governing this
 * kind of document.
 * <P>
 * If the attribute has not been explicitly assigned a value, but has
 * been declared in the DTD, it will exist and have that default. Only
 * if neither the document nor the DTD specifies a value will the
 * Attribute really be considered absent and have no value; in that
 * case, querying the attribute will return null.
 * <P>
 * Attributes may have multiple children that contain their data. (XML
 * allows attributes to contain entity references, and tokenized
 * attribute types such as NMTOKENS may have a child for each token.)
 * For convenience, the Attribute object's getValue() method returns
 * the string version of the attribute's value.
 * <P>
 * Attributes are not children of the Elements they belong to, in the
 * usual sense, and have no valid Parent reference. However, the spec
 * says they _do_ belong to a specific Element, and an INUSE exception
 * is to be thrown if the user attempts to explicitly share them
 * between elements.
 * <P>
 * Note that Elements do not permit attributes to appear to be shared
 * (see the INUSE exception), so this object's mutability is
 * officially not an issue.
 * 
 * @version Revision: 50 1.2 src/com/ibm/xml/dom/AttrImpl.java, parser, xml4j2, xml4j2_0_6 
 * @since  PR-DOM-Level-1-19980818.
 */
public class AttrImpl 
    extends NodeImpl 
    implements Attr {

    //
    // Data
    //

    /** Flag used for INUSE exception processing. */
    protected boolean owned;

    /** False for default attributes. */
    private boolean specified = true;

    //
    // Constructors
    //

    /**
     * This is the Fast constructor. Only the fNodeIndex is given here. 
     * All other data, can be requested from the ownerDocument via the index.
     */
    AttrImpl(DocumentImpl ownerDocument, int nodeIndex) {
        super(ownerDocument, nodeIndex);
    	this.specified = super.ownerDocument.getIntNodeValue(nodeIndex) == 1;        
    }

    /** 
     * Attribute has no public constructor. Please use the factory
     * method in the Document class.  
     */
    public AttrImpl(DocumentImpl ownerDocument, String name) {
    	super(ownerDocument, name, null);
    }

    //
    // Node methods
    //

    /** 
     * A short integer indicating what type of node this is. The named
     * constants for this value are defined in the org.w3c.dom.Node interface.
     */
    public final short getNodeType() {
        return Node.ATTRIBUTE_NODE;
    }

    /** 
     * Implicit in the rerouting of getNodeValue to getValue is the
     * need to redefine setNodeValue, for symmetry's sake.  Note that
     * since we're explicitly providing a value, Specified should be set
     * true.... even if that value equals the default.
     */
    public final void setNodeValue(String value) throws DOMException {
    	setValue(value);
    }

    /** 
     * In Attribute objects, NodeValue is considered a synonym for
     * Value.
     *
     * @see getValue() 
     */
    public final String getNodeValue() {
    	return getValue();
    }

    //
    // Attr methods
    //

    /** 
     * In Attributes, NodeName is considered a synonym for the
     * attribute's Name 
     */
    public final String getName() {

        if (syncData) {
            synchronizeData();
        }
    	return name;

    } // getName():String

    /** 
     * The DOM doesn't clearly define what setValue(null) means. I've taken it
     * as "remove all children", which from outside should appear
     * similar to setting it to the empty string.
     */
    public void setValue(String value) {

    	if (readOnly) {
    		throw new DOMExceptionImpl(
    			DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
        }
    		
        // Remove existing kids
        firstChild   = null;
        lastChild    = null;
        syncChildren = false;

        // Create and add the new one
    	if (value != null) {
    		appendChild(ownerDocument.createTextNode(value));
        }

    	specified = true;
    	changed();

    } // setValue(String)

    /**
     * The "string value" of an Attribute is its text representation,
     * which in turn is a concatenation of the string values of its children.
     */
    public String getValue() {

    	StringBuffer value = new StringBuffer();
    	Node node = getFirstChild();
    	while (node != null) {
    		value.append(node.getNodeValue());
    		node = node.getNextSibling();
    	}
    	return value.toString();

    } // getValue():String

    /**
     * The "specified" flag is true if and only if this attribute's
     * value was explicitly specified in the original document. Note that
     * the implementation, not the user, is in charge of this
     * property. If the user asserts an Attribute value (even if it ends
     * up having the same value as the default), it is considered a
     * specified attribute. If you really want to revert to the default,
     * delete the attribute from the Element, and the Implementation will
     * re-assert the default (if any) in its place, with the appropriate
     * specified=false setting.
     */
    public boolean getSpecified() {
    	return specified;
    }

    //
    // Public methods
    //

    /** NON-DOM, for use by parser */
    public void setSpecified(boolean arg) {
    	specified = arg;
    }

    //
    // Object methods
    //

    /** NON-DOM method for debugging convenience */
    public String toString() {
    	return getName() + "=" + "\"" + getValue() + "\"";
    }

} // class AttrImpl
