/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import java.util.Vector;

import com.ibm.xml.framework.AttrPool;
import com.ibm.xml.framework.EntityPool;
import com.ibm.xml.framework.ParserState;
import com.ibm.xml.framework.StringPool;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DocumentType;
import org.w3c.dom.DOMException;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Element;
import org.w3c.dom.Entity;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Notation;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

/**
 * The Document interface represents the entire HTML or XML document.
 * Conceptually, it is the root of the document tree, and provides the
 * primary access to the document's data.
 * <P>
 * Since elements, text nodes, comments, processing instructions,
 * etc. cannot exist outside the context of a Document, the Document
 * interface also contains the factory methods needed to create these
 * objects. The Node objects created have a ownerDocument attribute
 * which associates them with the Document within whose context they
 * were created.
 *
 * @version 56 1.6 src/com/ibm/xml/dom/DocumentImpl.java, parser, xml4j2, xml4j2_0_6 
 * @since  PR-DOM-Level-1-19980818.
 */
public final class DocumentImpl 
    extends NodeImpl 
    implements Document {

    //
    // Constants
    //

    // debugging

    /** To include code for printing the internal tables. */
    private static final boolean DEBUG_PRINT_TABLES = false;

    // private

    /** Chunk shift. */
    private static final int CHUNK_SHIFT = 11;           // 2^11 = 2k

    /** Chunk size. */
    private static final int CHUNK_SIZE = (1 << CHUNK_SHIFT);

    /** Chunk mask. */
    private static final int CHUNK_MASK = CHUNK_SIZE - 1;

    /** Initial chunk size. */
    private static final int INITIAL_CHUNK_COUNT = (1 << (16 - CHUNK_SHIFT));   // 2^16 = 64k
    
    //
    // Data
    //

    // lazy-eval information

    /** Node count. */
    private int fNodeCount = 0;
    
    /** Node types. */
    private byte[][] fNodeType;

    /** Node names. */
    private int[][] fNodeName;
    
    /** Node values. */
    private int[][] fNodeValue;

    /** Node parents. */
    private int[][] fNodeParent;

    /** Node first children. */
    private int[][] fNodeFirstChild;

    /** Node last children. */
    private int[][] fNodeLastChild;

    /** Node previous siblings. */
    private int[][] fNodePrevSib;

    /** Node next siblings. */
    private int[][] fNodeNextSib;

    // document information

    /** Document type. */
    private DocumentTypeImpl docType;

    /** Document element. */
    private ElementImpl docElement;

    // parser pools

    /** Parser state. */
    protected ParserState fParserState;

    // local cache

    /** String pool cache. */
    private StringPool fStringPool;

    //
    // Constructors
    //

    /** 
     * NON-DOM: Actually creating a Document is outside the DOM's spec,
     * since it has to operate in terms of a particular implementation.
     */
    public DocumentImpl() {
        super(null, null, null);
    }
    
    /** 
     * NON-DOM: Actually creating a Document is outside the DOM's spec,
     * since it has to operate in terms of a particular implementation.
     */
    public DocumentImpl(ParserState state) {
        super(null, 0);

        fParserState = state;
        fStringPool  = state.getStringPool();
        
    } // <init>(ParserState)

    //
    // Node methods
    //

    /** Returns the node type. */
    public final short getNodeType() {
        return Node.DOCUMENT_NODE;
    }

    /** Returns the node name. */
    public final String getNodeName() {
        return "#document";
    }

    /** Returns the node value. */
    public final String getNodeValue() {
        return null;
    }

    /**
     * Deep-clone a document, including fixing ownerDoc for the cloned
     * children. Note that this requires bypassing the WRONG_DOCUMENT_ERR
     * protection. I've chosen to implement it by calling importNode --
     * which is NON-DOM in Level 1, but which anticipates one likely 
     * direction the Level 2 DOM might take in solving this.
     * 
     * @return org.w3c.dom.Node
     * @param deep boolean, iff true replicate children
     */
    public Node cloneNode(boolean deep) {

    	DocumentImpl newdoc = fNodeIndex == 0 ?
    	                        new DocumentImpl(fParserState) :
    	                        new DocumentImpl();
    	if (deep) {
    		for(NodeImpl n = (NodeImpl)getFirstChild(); 
                n != null; 
                n = n.nextSibling) 
                {
    			newdoc.appendChild(newdoc.importNode(n,true));
            }
        }

    	return newdoc;

    } // cloneNode(boolean):Node

    /**
     * Since a Document may contain at most one top-level Element child,
  	 * and at most one DocumentType declaraction, we need to subclass our
  	 * add-children methods to implement this constraint. 
	 * Since appendChild() is implemented as insertBefore(,null),
	 * altering the latter fixes both.
  	 * <p>
  	 * While I'm doing so, I've taken advantage of the opportunity to
  	 * cache documentElement and docType so we don't have to
  	 * search for them.
	 */
    public Node insertBefore(Node newChild, Node refChild) 
        throws DOMException {

    	// Only one such child permitted
        int type = newChild.getNodeType();
    	if((type == Node.ELEMENT_NODE && docElement != null) ||
    	   (type == Node.DOCUMENT_TYPE_NODE && docType != null)) {
    	    throw new DOMExceptionImpl(DOMException.HIERARCHY_REQUEST_ERR,null);
        }
    
    	super.insertBefore(newChild,refChild);
    
    	// If insert succeeded, cache the kid appropriately
        if (type == Node.ELEMENT_NODE) {
    	    docElement = (ElementImpl)newChild;
        }
        else if (type == Node.DOCUMENT_TYPE_NODE) {
    	    docType=(DocumentTypeImpl)newChild;
        }
    
    	return newChild;

    } // insertBefore(Node,Node):Node

    /** 
     * Since insertBefore caches the docElement (and, currently, docType),
	 * removeChild has to know how to undo the cache 
     */
    public Node removeChild(Node oldChild) 
        throws DOMException {
	    super.removeChild(oldChild);
	
    	// If remove succeeded, un-cache the kid appropriately
        int type = oldChild.getNodeType();
        if(type == Node.ELEMENT_NODE) {
    	    docElement = null;
        }
        else if (type == Node.DOCUMENT_TYPE_NODE) {
    	    docType=null;
        }
    
    	return oldChild;

    }   // removeChild(Node):Node

    /**
     * Documents never have a nodeValue.
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR)
     */
    public void setNodeValue(String x) 
        throws DOMException {
    	throw new DOMExceptionImpl(DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
    }

    //
    // Document methods
    //

    // factory methods

    /**
     * Factory method; creates an Attribute having this Document as its
	 * OwnerDoc.
     *
	 * @param name The name of the attribute. Note that the attribute's value
	 * is _not_ established at the factory; remember to set it!
     *
	 * @throws DOMException(INVALID_NAME_ERR) if the attribute name is not
	 * acceptable. 
     */
    public Attr createAttribute(String name) 
        throws DOMException {

    	if(!isXMLName(name)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR,null);
        }

    	return new AttrImpl(this, name); 

    } // createAttribute(String):Attr

    /**
     * Factory method; creates a CDATASection having this Document as
	 * its OwnerDoc.
     *
	 * @param data The initial contents of the CDATA
     *
	 * @throws DOMException(NOT_SUPPORTED_ERR) for HTML documents. (HTML
	 * not yet implemented.)
	 */
    public CDATASection createCDATASection(String data)
	    throws DOMException {
	    return new CDATASectionImpl(this, data);
    }  

    /**
     * Factory method; creates a Comment having this Document as its
     * OwnerDoc.
     *
     * @param data The initial contents of the Comment. */
    public Comment createComment(String data) {
	    return new CommentImpl(this, data);
    }  

    /**
     * Factory method; creates a DocumentFragment having this Document
	 * as its OwnerDoc. 
     */
    public DocumentFragment createDocumentFragment() {
	    return new DocumentFragmentImpl(this);
    }  

    /** 
     * Factory method; creates an Element having this Document
	 * as its OwnerDoc.
     *
	 * @param tagName The name of the element type to instantiate. For
	 * XML, this is case-sensitive. For HTML, the tagName parameter may
	 * be provided in any case, but it must be mapped to the canonical
	 * uppercase form by the DOM implementation.
     *
	 * @throws DOMException(INVALID_NAME_ERR) if the tag name is not
	 * acceptable.
	 */
    public Element createElement(String tagName) 
        throws DOMException {

    	if (!isXMLName(tagName)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR, null);
        }                           
    
    	return new ElementImpl(this, tagName);

    } // createElement(String):Element

    /**
     * Factory method; creates an EntityReference having this Document
     * as its OwnerDoc.
     * 
     * @param name The name of the Entity we wish to refer to
     * 
     * @throws DOMException(NOT_SUPPORTED_ERR) for HTML documents, where
     * nonstandard entities are not permitted. (HTML not yet
     * implemented.)  
     */
    public EntityReference createEntityReference(String name) 
        throws DOMException {

    	if (!isXMLName(name)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR, null);
        }
    
    	return new EntityReferenceImpl(this, name);

    } // createEntityReference(String):EntityReference

    /** 
     * Factory method; creates a ProcessingInstruction having this Document
	 * as its OwnerDoc.
     *
	 * @param target The target "processor channel"
	 * @param data Parameter string to be passed to the target.
     *
	 * @throws DOMException(INVALID_NAME_ERR) if the target name is not
	 * acceptable.
     *
	 * @throws DOMException(NOT_SUPPORTED_ERR) for HTML documents. (HTML
	 * not yet implemented.)
	 */
    public ProcessingInstruction createProcessingInstruction(String target, String data)
        throws DOMException {

    	if(!isXMLName(target)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR,null);
        }

    	return new ProcessingInstructionImpl(this, target, data);

    } // createProcessingInstruction(String,String):ProcessingInstruction

    /** 
     * Factory method; creates a Text node having this Document as its
	 * OwnerDoc.
     *
	 * @param data The initial contents of the Text.  
     */
    public Text createTextNode(String data) {
        return new TextImpl(this, data);
    }  

    // other document methods

    /**
     * For XML, this provides access to the Document Type Definition.
	 * For HTML documents, and XML documents which don't specify a DTD,
	 * it will be null.
	 */
    public DocumentType getDoctype() {
        if (syncChildren) {
            synchronizeChildren();
        }
	    return docType;
    }  

    /**
     * Convenience method, allowing direct access to the child node
	 * which is considered the root of the actual document content. For
	 * HTML, where it is legal to have more than one Element at the top
	 * level of the document, we pick the one with the tagName
	 * "HTML". For XML there should be only one top-level
     *
	 * (HTML not yet supported.)  
     */
    public Element getDocumentElement() {
        if (syncChildren) { 
            synchronizeChildren(); 
        }
	    return docElement;
    }  

    /**
     * Return a <em>live</em> collection of all descendent Elements (not just
	 * immediate children) having the specified tag name.
     *
	 * @param tagname The type of Element we want to gather. "*" will be
	 * taken as a wildcard, meaning "all elements in the document."
     * 
	 * @see com.ibm.domimpl.DeepNodeListImpl
	 */
    public NodeList getElementsByTagName(String tagname) {
	    return new DeepNodeListImpl(this,tagname);
    }  

    /**
     * Retrieve information describing the abilities of this particular
	 * DOM implementation. Intended to support applications that may be
	 * using DOMs retrieved from several different sources, potentially
	 * with different underlying representations.
	 */
    public DOMImplementation getImplementation() {
        // Currently implemented as a singleton, since it's hardcoded
        // information anyway.
        return DOMImplementationImpl.getDOMImplementation();
    }  

    //
    // Public methods
    //

    // non-DOM factory methods

    /**
     * NON-DOM
     * Factory method; creates a DocumentType having this Document
     * as its OwnerDoc. (REC-DOM-Level-1-19981001 left the process of building
     * DTD information unspecified.)
     * 
     * @param name The name of the Entity we wish to provide a value for.
     * 
     * @throws DOMException(NOT_SUPPORTED_ERR) for HTML documents, where
     * DTDs are not permitted. (HTML not yet implemented.)  
     */
    public DocumentType createDocumentType(String name) 
        throws DOMException {

    	if (!isXMLName(name)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR, null);
        }
    
    	return new DocumentTypeImpl(this, name);

    } // createDocumentType(String):DocumentType

    /**
     * NON-DOM
     * Factory method; creates an Entity having this Document
     * as its OwnerDoc. (REC-DOM-Level-1-19981001 left the process of building
     * DTD information unspecified.)
     * 
     * @param name The name of the Entity we wish to provide a value for.
     * 
     * @throws DOMException(NOT_SUPPORTED_ERR) for HTML documents, where
     * nonstandard entities are not permitted. (HTML not yet
     * implemented.)  
     */
    public Entity createEntity(String name) 
        throws DOMException {

        // REVISIT: Should we be checking XML name chars?
        /***
    	if (!isXMLName(name)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR, null);
        }
        /***/
    
    	return new EntityImpl(this, name);

    } // createEntity(String):Entity

    /**
     * NON-DOM
     * Factory method; creates a Notation having this Document
     * as its OwnerDoc. (REC-DOM-Level-1-19981001 left the process of building
     * DTD information unspecified.)
     * 
     * @param name The name of the Notation we wish to describe
     * 
     * @throws DOMException(NOT_SUPPORTED_ERR) for HTML documents, where
     * notations are not permitted. (HTML not yet
     * implemented.)  
     */
    public Notation createNotation(String name) 
        throws DOMException {

        // REVISIT: Should we be checking XML name chars?
        /***
    	if (!isXMLName(name)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR, null);
        }                       
        /***/
    
    	return new NotationImpl(this, name);

    } // createNotation(String):Notation

    /**
     * NON-DOM Factory method: creates an element definition. Element
     * definitions hold default attribute values.
     */
    public ElementDefinitionImpl createElementDefinition(String name) 
        throws DOMException {

        // REVISIT: Should we be checking XML name chars?
        /***
    	if (!isXMLName(name)) {
    		throw new DOMExceptionImpl(DOMException.INVALID_CHARACTER_ERR, null);
        }
        /***/
    
        return new ElementDefinitionImpl(this, name);

    } // createElementDefinition(String):ElementDefinitionImpl

    // other non-DOM methods

    /**
     * NON-DOM: Copies data from the source node. Unlike cloneNode, this
     * _can_ copy data from another document. If the source document is also
     * based on com.ibm.domimpl, we will attempt to preserve the domimpl-
     * internal data by doing a clone-and-reparent. If not, we will use
     * the source's public methods, and this document's Factory methods,
     * to copy data defined by the DOM interfaces.
     *
     * Its behavior is otherwise similar to that of cloneNode.
     *
     * Attempting to import a Document into another Document is meaningless --
     * a new Document would not improve matters much, and a DocumentFragment
     * couldn't carry the DocumentType child (if any). Best thing we can do
     * is throw a HIERARCHY_REQUEST_ERR.
     *
     * ????? Should we push some of this down to copy-ctors, so
     * subclassed DOMs have the option of special-casing each other
     * (as we do for ourself)?
     */
    public Node importNode(Node source, boolean deep) 
        throws DOMException {

    	NodeImpl newnode=null;
    
    	// Sigh. This doesn't work; too many nodes have private data that
    	// would have to be manually tweaked. May be able to add local
    	// shortcuts to each nodetype. Consider ?????
    	//if(source instanceof NodeImpl && 
    	//	!(source instanceof DocumentImpl))
    	//{
    	//  // Can't clone DocumentImpl since it invokes us...
    	//	newnode=(NodeImpl)source.cloneNode(false);
    	//	newnode.ownerDocument=this;
    	//}
    	//else 
        int type = source.getNodeType();
    	switch (type) {
    		
            case ELEMENT_NODE: {
    			Element newelement = createElement(source.getNodeName());
    			NamedNodeMap srcattr = source.getAttributes();
    			if (srcattr != null) {
    				for(int i = 0; i < srcattr.getLength(); i++) {
    					newelement.setAttributeNode(
    						(AttrImpl)importNode(srcattr.item(i), true));
                    }
                }
    			newnode = (NodeImpl)newelement;
    			break;
            }

            case ATTRIBUTE_NODE: {
    			newnode = (NodeImpl)createAttribute(source.getNodeName());
    			// Kids carry value
    			break;
            }

    		case TEXT_NODE: {
    			newnode = (NodeImpl)createTextNode(source.getNodeValue());
    			break;
            }

    		case CDATA_SECTION_NODE: {
    			newnode = (NodeImpl)createCDATASection(source.getNodeValue());
    			break;
            }

    		case ENTITY_REFERENCE_NODE: { 
    			newnode = (NodeImpl)createEntityReference(source.getNodeName());
    			deep = false; // ????? Right Thing?
    			// Value implied by doctype, so we should not copy it
    			// -- instead, refer to local doctype, if any.
    			break;
            }

    		case ENTITY_NODE: {
    			Entity srcentity = (Entity)source;
    			EntityImpl newentity = (EntityImpl)createEntity(source.getNodeName());
    			newentity.setPublicId(srcentity.getPublicId());
    			newentity.setSystemId(srcentity.getSystemId());
    			newentity.setNotationName(srcentity.getNotationName());
    			// Kids carry additional value
    			newnode = newentity;
    			break;  
            }

    		case PROCESSING_INSTRUCTION_NODE: {
    			newnode = (ProcessingInstructionImpl)createProcessingInstruction(source.getNodeName(), source.getNodeValue());
    			break;
            }

    		case COMMENT_NODE: {
    			newnode = (NodeImpl)createComment(source.getNodeValue());
    			break;
            }

    		case DOCUMENT_TYPE_NODE: {
    			DocumentTypeImpl newdoctype = (DocumentTypeImpl)createDocumentType(source.getNodeName());
    			// Values are on NamedNodeMaps
    			NamedNodeMap smap = ((DocumentType)source).getEntities();
    			NamedNodeMap tmap = newdoctype.getEntities();
    			if(smap != null) {
    				for(int i = 0; i < smap.getLength(); i++) {
    					tmap.setNamedItem((EntityImpl)importNode(smap.item(i), true));
                    }
                }
    			smap = ((DocumentType)source).getNotations();
    			tmap = newdoctype.getNotations();
    			if (smap != null) {
    				for(int i = 0; i < smap.getLength(); i++) {
    					tmap.setNamedItem((NotationImpl)importNode(smap.item(i), true));
                    }
                }
    			// NOTE: At this time, the DOM definition of DocumentType
    			// doesn't cover Elements and their Attributes. domimpl's
    			// extentions in that area will not be preserved, even if
    			// copying from domimpl to domimpl. We could special-case
    			// that here. Arguably we should. Consider. ?????
    		    newnode = newdoctype;
    			break;
            }

    		case DOCUMENT_FRAGMENT_NODE: {
    			newnode = (NodeImpl)createDocumentFragment();
    			// No name, kids carry value
    			break;
            }

    		case NOTATION_NODE: {
    			Notation srcnotation = (Notation)source;
    			NotationImpl newnotation = (NotationImpl)createNotation(source.getNodeName());
    			newnotation.setPublicId(srcnotation.getPublicId());
    			newnotation.setSystemId(srcnotation.getSystemId());
    			// Kids carry additional value
    			newnode = newnotation;
    			// No name, no value
    			break;
            }
    
    		case DOCUMENT_NODE : // Document can't be child of Document
    		default: {			 // Unknown node type
    			throw new DOMExceptionImpl(DOMException.HIERARCHY_REQUEST_ERR,null);
            }
        }
    
    	// If deep, replicate and attach the kids.
    	if (deep) {
    		for (Node srckid = source.getFirstChild(); 
                 srckid != null; 
                 srckid = srckid.getNextSibling()) {
    			newnode.appendChild(importNode(srckid, true));
    		}
        }
    
    	return newnode;

    } // importNode(Node,boolean):Node

    // internal factory methods

    /** Creates a document node in the table. */
    public final int createDocument() {
        int nodeIndex = createNode(Node.DOCUMENT_NODE);
        return nodeIndex;
    }

    /** Creates a doctype. */
    public final int createDocumentType(int rootElementNameIndex) {
        int nodeIndex = createNode(Node.DOCUMENT_TYPE_NODE);
        int chunk     = nodeIndex >> CHUNK_SHIFT;
        int index     = nodeIndex & CHUNK_MASK;
        fNodeName[chunk][index] = rootElementNameIndex;
        return nodeIndex;
    }

    /** Creates a notation in the table. */
    public final int createNotation(int notationIndex) {

        // make entity node
        int nodeIndex = createNode(Node.NOTATION_NODE);
        int chunk     = nodeIndex >> CHUNK_SHIFT;
        int index     = nodeIndex & CHUNK_MASK;

        // make extra data node 
        int extraDataIndex = createNode(Node.TEXT_NODE);
        int echunk = extraDataIndex >> CHUNK_SHIFT;
        int eindex = extraDataIndex & CHUNK_MASK;

        // save name, public id, system id, and notation name
        fNodeValue[chunk][index] = extraDataIndex;
        EntityPool pool = fParserState.getEntityPool();
        fNodeName[chunk][index]         = pool.getNotationName(notationIndex);
        fNodeFirstChild[echunk][eindex] = pool.getPublicId(notationIndex);
        fNodeLastChild[echunk][eindex]  = pool.getSystemId(notationIndex);

        return nodeIndex;

    } // createNotation(int):int

    /** Creates an entity in the table. */
    public final int createEntity(int entityIndex) {

        // make entity node
        int nodeIndex = createNode(Node.ENTITY_NODE);
        int chunk     = nodeIndex >> CHUNK_SHIFT;
        int index     = nodeIndex & CHUNK_MASK;

        // make extra data node 
        int extraDataIndex = createNode(Node.TEXT_NODE);
        int echunk = extraDataIndex >> CHUNK_SHIFT;
        int eindex = extraDataIndex & CHUNK_MASK;

        // save name, public id, system id, and notation name
        fNodeValue[chunk][index] = extraDataIndex;
        EntityPool pool = fParserState.getEntityPool();
        fNodeName[chunk][index]         = pool.getEntityName(entityIndex);
        fNodeFirstChild[echunk][eindex] = pool.getPublicId(entityIndex);
        fNodeLastChild[echunk][eindex]  = pool.getSystemId(entityIndex);
        fNodePrevSib[echunk][eindex]    = pool.getNotationName(entityIndex);

        return nodeIndex;

    } // createEntity(int):int

    /** Creates an entity reference node in the table. */
    public final int createEntityReference(int entityRefIndex) {

        int nodeIndex = createNode(Node.ENTITY_REFERENCE_NODE);
        int chunk     = nodeIndex >> CHUNK_SHIFT;
        int index     = nodeIndex & CHUNK_MASK;

        // save name
        int nameIndex = fParserState.getEntityPool().getEntityName(entityRefIndex);
        fNodeName[chunk][index] = nameIndex;
        return nodeIndex;

    } // createEntityReference(int):int

    /** Creates an element node in the table. */
    public final int createElement(int elementNameIndex, int attrListIndex) {

        // create element
        int elementNodeIndex = createNode(Node.ELEMENT_NODE);
        int elementChunk     = elementNodeIndex >> CHUNK_SHIFT;
        int elementIndex     = elementNodeIndex & CHUNK_MASK;
        fNodeName[elementChunk][elementIndex] = elementNameIndex;
        //fNodeValue[chunk][index] = attrListIndex;

        // create attributes
        if (attrListIndex != -1) {
            AttrPool attrPool = fParserState.getAttrPool();
            int first = attrPool.getFirstAttr(attrListIndex);
            int lastAttrNodeIndex = -1;
            int lastAttrChunk = -1;
            int lastAttrIndex = -1;
            for (int index = first; index != -1; 
                 index = attrPool.getNextAttr(index)) {

                // REVISIT: Replace most of this next code with a call
                //          to createAttribute(). -Ac

                // create attribute
                int attrNodeIndex = createNode(Node.ATTRIBUTE_NODE);
                int attrChunk = attrNodeIndex >> CHUNK_SHIFT;
                int attrIndex  = attrNodeIndex & CHUNK_MASK;
                fNodeParent[attrChunk][attrIndex] = elementNodeIndex;
                fNodeName[attrChunk][attrIndex]   = attrPool.getAttrName(index);
                fNodeValue[attrChunk][attrIndex]  = 
                    attrPool.isSpecified(index) ? 1 : 0;

                int textNodeIndex = createNode(Node.TEXT_NODE);
                int textChunk = textNodeIndex >> CHUNK_SHIFT;
                int textIndex = textNodeIndex & CHUNK_MASK;
                fNodeValue[textChunk][textIndex] = attrPool.getAttValue(index);
                appendChild(attrNodeIndex, textNodeIndex);

                // add links
                if (index == first) {
                    fNodeValue[elementChunk][elementIndex] = attrNodeIndex;
                }
                else {
                    fNodeNextSib[lastAttrChunk][lastAttrIndex] = attrNodeIndex;
                    fNodePrevSib[attrChunk][attrIndex] = lastAttrNodeIndex;
                }

                // save last chunk and index
                lastAttrNodeIndex = attrNodeIndex;
                lastAttrChunk     = attrChunk;
                lastAttrIndex     = attrIndex;
            }
        }

        return elementNodeIndex;

    } // createElement(int,int):int

    /** Creates an attributes in the table. */
    public final int createAttribute(int attrNameIndex, int attrValueIndex,
                                     boolean specified) {

        // create attribute and set specified
        int nodeIndex = createNode(NodeImpl.ATTRIBUTE_NODE);
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        fNodeName[chunk][index]  = attrNameIndex;
        fNodeValue[chunk][index] = specified ? 1 : 0;
        
        // append value as text node
        int textNodeIndex = createTextNode(attrValueIndex, false);
        appendChild(nodeIndex, textNodeIndex);

        // return attribute node
        return nodeIndex;

    } // createAttribute(int):int

    /** Creates an element definition in the table. */
    public final int createElementDefinition(int elementNameIndex) {

        int nodeIndex = createNode(NodeImpl.ELEMENT_DEFINITION_NODE);
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        fNodeName[chunk][index] = elementNameIndex;

        return nodeIndex;

    } // createElementDefinition(int):int

    /** Creates a text node in the table. */
    public final int createTextNode(int dataIndex, boolean ignorableWhitespace) {

        int nodeIndex = createNode(Node.TEXT_NODE);
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        fNodeValue[chunk][index] = dataIndex;
        fNodeFirstChild[chunk][index] = ignorableWhitespace ?  1 : 0;
        return nodeIndex;

    } // createTextNode(int):int

    /** Creates a CDATA section node in the table. */
    public final int createCDATASection(int dataIndex) {

        int nodeIndex = createNode(Node.CDATA_SECTION_NODE);
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        fNodeValue[chunk][index] = dataIndex;

        return nodeIndex;

    } // createCDATASection(int):int

    /** Creates a processing instruction node in the table. */
    public final int createProcessingInstruction(int targetIndex, int dataIndex) {

        int nodeIndex = createNode(Node.PROCESSING_INSTRUCTION_NODE);
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        fNodeName[chunk][index] = targetIndex;
        fNodeValue[chunk][index] = dataIndex;

        return nodeIndex;

    } // createProcessingInstruction(int,int):int

    /** Creates a comment node in the table. */
    public final int createComment(int dataIndex) {

        int nodeIndex = createNode(Node.COMMENT_NODE);
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        fNodeValue[chunk][index] = dataIndex;

        return nodeIndex;

    } // createComment(int):int

    /** Appends a child to the specified parent in the table. */
    public final void appendChild(int parentIndex, int childIndex) {

        int pchunk = parentIndex >> CHUNK_SHIFT;
        int pindex = parentIndex & CHUNK_MASK;
        int chunk = childIndex >> CHUNK_SHIFT;
        int index = childIndex & CHUNK_MASK;
        fNodeParent[chunk][index] = parentIndex;
        int prev = fNodeLastChild[pchunk][pindex];
        fNodePrevSib[chunk][index] = prev;
        if (prev == -1) {
            fNodeFirstChild[pchunk][pindex] = childIndex;
            }
        else {
            int chnk = prev >> CHUNK_SHIFT;
            int indx = prev & CHUNK_MASK;
            fNodeNextSib[chnk][indx] = childIndex;
        }
        fNodeLastChild[pchunk][pindex] = childIndex;

    } // appendChild(int,int)
    
    /** Sets the first child of the parentIndex to childIndex. */
    public final void setAsFirstChild(int parentIndex, int childIndex) {

        int pchunk = parentIndex >> CHUNK_SHIFT;
        int pindex = parentIndex & CHUNK_MASK;
        int chunk = childIndex >> CHUNK_SHIFT;
        int index = childIndex & CHUNK_MASK;
        fNodeFirstChild[pchunk][pindex] = childIndex;   
        
        int next = childIndex;
        
        while (next != -1) {
                
            childIndex = next;
            next = fNodeNextSib[chunk][index];
            chunk = next >> CHUNK_SHIFT;
            index = next & CHUNK_MASK;
        }
        
        fNodeLastChild[pchunk][pindex] = childIndex;

    } // setAsFirstChild(int,int)
        
    // methods used when objects are "fluffed-up"

    /** Returns the parent node of the given node. */
    public final int getParentNode(int nodeIndex) {

        // REVISIT: should this be "public"? -Ac

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        return fNodeParent[chunk][index];

    } // getParentNode(int):int

    /** Returns the first child of the given node. */
    public final int getFirstChild(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        return fNodeFirstChild[chunk][index];

    } // getFirstChild(int):int

    /** Returns the last child of the given node. */
    public final int getLastChild(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        int lastChild =  fNodeLastChild[chunk][index];
        
        // revisit: the following could be more efficient.
 
        // if the last child is a TEXT_NODE
        if (lastChild != -1 && fNodeType[chunk][index] == Node.TEXT_NODE) {
            // get the previous sibling.
            int previousIndex = fNodePrevSib[chunk][index];
            chunk = previousIndex >> CHUNK_SHIFT;
            index = previousIndex & CHUNK_MASK;
            // if the previous Sibling is a TEXT_NODE
            if (previousIndex != -1 && fNodeType[chunk][index] == Node.TEXT_NODE) {   
                // while previous sibling is a text node.
                while (previousIndex != -1 && fNodeType[chunk][index] == Node.TEXT_NODE) {
                    nodeIndex = previousIndex;
                    previousIndex = fNodePrevSib[chunk][index];
                    chunk = previousIndex >> CHUNK_SHIFT;
                    index = previousIndex & CHUNK_MASK; 
                }
                return nodeIndex;        
            } 
        }

        return lastChild;

    } // getLastChild(int):int

    /** Returns the previous sibling of the given node. */
    public final int getPreviousSibling(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        int previousIndex = fNodePrevSib[chunk][index];
        
        // revisit: the following could be more efficient.
        
        // if the previous sibling exists and this node is not a text node.
        if (previousIndex != -1 
        && fNodeType[chunk][index] != Node.TEXT_NODE) {
            chunk = previousIndex >> CHUNK_SHIFT;
            index = previousIndex & CHUNK_MASK;
            // if the previous node is a text node.
            if (fNodeType[chunk][index] == Node.TEXT_NODE) {      
                // while the previous node exists and is a text node.
                while (previousIndex != -1 
                    && fNodeType[chunk][index] == Node.TEXT_NODE) {
                    nodeIndex = previousIndex;
                    previousIndex = fNodePrevSib[chunk][index];
                    chunk = previousIndex >> CHUNK_SHIFT;
                    index = previousIndex & CHUNK_MASK; 
                }
                return nodeIndex;        
            } 
        }

        return previousIndex;

    } // getPreviousSibling(int):int

    /** Returns the next sibling of the given node. 
     *  This is post-normalization of Text Nodes.
     */
    public final int getNextSibling(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        nodeIndex = fNodeNextSib[chunk][index];
        
        while (nodeIndex != -1 && fNodeType[chunk][index] == Node.TEXT_NODE) {
            nodeIndex = fNodeNextSib[chunk][index];
            chunk = nodeIndex >> CHUNK_SHIFT;
            index = nodeIndex & CHUNK_MASK;
        }

        return nodeIndex;

    } // getNextSibling(int):int

    /** 
     * Returns the <i>real</i> next sibling of the given node, 
     * directly from the data structures. Used by TextImpl#getNodeValue()
     * to normalize values.
     */
    public final int getRealNextSibling(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        return fNodeNextSib[chunk][index];

        } // getReadNextSibling(int):int
 
    /** 
     * Returns the index of the element definition in the table
     * with the specified name index, or -1 if no such definition
     * exists.
     */
    public final int lookupElementDefinition(int elementNameIndex) {

        if (fNodeCount > 1) {

            // find doctype
            int docTypeIndex = -1;
            for (int index = getFirstChild(0); // 0 == document node
                 index != -1; 
                 index = getNextSibling(index)) {
                if (getNodeType(index) == Node.DOCUMENT_TYPE_NODE) {
                    docTypeIndex = index;
                    break;
                }
            }

            // find element definition
            for (int index = getFirstChild(docTypeIndex); 
                 index != -1; 
                 index = getNextSibling(index)) {
                if (getIntNodeName(index) == elementNameIndex) {
                    return index;
                }
            }
        }

        return -1;

    } // lookupElementDefinition(int):int

    /** Returns the attribute list index for elements. */
    public final int getAttributeList(int elementNodeIndex) {

        if (elementNodeIndex == -1) {
            return -1;
        }

        int chunk = elementNodeIndex >> CHUNK_SHIFT;
        int index = elementNodeIndex & CHUNK_MASK;
        return fNodeValue[chunk][index];

    } // getAttributeList(int):int

    /** Instantiates the requested node object. */
    public final NodeImpl getNodeObject(int nodeIndex) {

        // is there anything to do?
        if (nodeIndex == -1) {
            return null;
        }

        // get node type
        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        int type = fNodeType[chunk][index];
        
        // create new node
        NodeImpl node = null;
        switch (type) {
            
            //
            // Standard DOM node types
            //

            case Node.ATTRIBUTE_NODE: {
                node = new AttrImpl(this, nodeIndex);
                break;
            }
            
            case Node.CDATA_SECTION_NODE: {
                node = new CDATASectionImpl(this, nodeIndex);
                break;
            }
            
            case Node.COMMENT_NODE: {
                node = new CommentImpl(this, nodeIndex);
                break;
            }
            
            // NOTE: Document fragments can never be "fast".
            //
            //       The parser will never ask to create a document
            //       fragment during the parse. Document fragments
            //       are used by the application *after* the parse.
            //
            // case Node.DOCUMENT_FRAGMENT_NODE: { break; }
            
            case Node.DOCUMENT_NODE: {
                // this node is never "fast"
                node = this;
                break;
            }
            
            case Node.DOCUMENT_TYPE_NODE: {
                node = new DocumentTypeImpl(this, nodeIndex);
                // Save away the doctype node
                docType = (DocumentTypeImpl)node;
                break;
            }
            
            case Node.ELEMENT_NODE: {
                node = new ElementImpl(this, nodeIndex);
                // Save away the document element node
                if (docElement == null) {
                    docElement = (ElementImpl)node;
                }
                break;
            }
            
            case Node.ENTITY_NODE: {
                node = new EntityImpl(this, nodeIndex);
                break;
            }
            
            case Node.ENTITY_REFERENCE_NODE: {
                node = new EntityReferenceImpl(this, nodeIndex);
                break;
            }
            
            case Node.NOTATION_NODE: {
                node = new NotationImpl(this, nodeIndex);
                break;
            }
            
            case Node.PROCESSING_INSTRUCTION_NODE: {
                node = new ProcessingInstructionImpl(this, nodeIndex);
                break;
            }
            
            case Node.TEXT_NODE: {
                node = new TextImpl(this, nodeIndex);
                break;
            }
            
            //
            // non-standard DOM node types
            //

            case NodeImpl.ELEMENT_DEFINITION_NODE: {
                node = new ElementDefinitionImpl(this, nodeIndex);
                break;
            }

        } // switch node type

        // store and return
        if (node != null) {
            return node;
        }

        // error
        throw new IllegalArgumentException();

    } // createNodeObject(int):Node

    /** Returns the name of the given node. */
    public final String getNodeName(int nodeIndex) {

        if (nodeIndex == -1) {
            return null;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        int nameIndex = fNodeName[chunk][index];
        if (nameIndex == -1) {
            return null;
        }

        return fStringPool.toString(nameIndex);

    } // getNodeName(int):String

    /** Returns the value of the given node. */
    public final String getNodeValue(int nodeIndex) {

        if (nodeIndex == -1) {
            return null;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        int valueIndex = fNodeValue[chunk][index];
        if (valueIndex == -1) {
            return null;
        }

        return fStringPool.toString(valueIndex);

    } // getNodeValue(int):String

    /** Returns the real int name of the given node. */
    public final int getIntNodeName(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        return fNodeName[chunk][index];

    } // getIntNodeName(int):int
  
    /** 
     * Returns the real int value of the given node. 
     *  Used by AttrImpl to store specified value (1 == true).
     */
    public final int getIntNodeValue(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        return fNodeValue[chunk][index];

    } // getIntNodeValue(int):int
  
    /** Returns the type of the given node. */
    public final short getNodeType(int nodeIndex) {

        if (nodeIndex == -1) {
            return -1;
        }

        int chunk = nodeIndex >> CHUNK_SHIFT;
        int index = nodeIndex & CHUNK_MASK;
        return fNodeType[chunk][index];

    } // getNodeType(int):int
    
    //
    // Public static methods
    //

    /**
     * Check the string against XML's definition of acceptable names for
     * elements and attributes and so on. From the XML spec:
     * <p>
     * [Definition:] A Name is a token beginning with a letter or one of a
     * few punctuation characters, and continuing with letters, digits,
     * hyphens,underscores, colons, or full stops, together known as name
     * characters.
     * <p>
     * Unfortunately, that spec goes on to say that after the first character,
     * names may use "combining characters" and "extender characters", 
     * which are explicitly listed rather than defined in terms of Java's
     * Unicode character types... and which in fact can't be expressed solely
     * in terms of those types.
     * <p>
     * I've empirically derived some tests which are partly based on the
     * Java Unicode space (which simplifies them considerably), but they
     * still wind up having to be further qualified. This may not remain
     * valid if future extensions of Java and/or Unicode introduce other
     * characters having these type numbers.
     * <p>
     * Suggestions for alternative implementations would be welcome.
     */
    public static boolean isXMLName(String s) {

        // REVISIT: Use the parser's character checking code. -Ac

    	char [] ca=new char[s.length()];
    	s.getChars(0,s.length(),ca,0);
    
    	// First character must be letter, underscore, or colon.
    	if (!Character.isLetter(ca[0]) && "_:".indexOf((int)ca[0]) == -1) {
    		return false;
            }
    
    	// Remaining characters must be letter, digit, underscore,
    	// colon, period, dash, an XML "Combining Character", or an
    	// XML "Extender Character".
    	for (int i = 1; i < s.length(); ++i) {

    		char c = ca[i];
    		int ctype = Character.getType(c);
    
    		if (!Character.isLetterOrDigit(c) && 
                (".-_:".indexOf(c) == -1) &&
    		  
                // Right type
    		    (!(ctype >= 6 && ctype <= 8 &&	
                      // Bad ranges, combined from the three types:
    				  !((c >= 0x06dd && c <= 0x06de) ||
      				  	(c >= 0x20dd && c <= 0x20e0) || 
      				   	 c >= 0x309b
    				   	)
    				  )
    			    ) &&
    		    
    			  // XML Extender chars are all type 4 (uppercase) except 
    			  // for two which are type 24. (titlecase modifier)

                  // Right type
    		      (!(ctype == 4 &&
                        // Bad ranges
    					!((c >= 0x02d0 && c <= 0x0559) ||
    				      (c >= 0x06e5 && c <= 0x06e6) ||
    					  (c >= 0x309b && c <= 0x309c)
    					 )
    				|| c == 0x00b7 // Type 24
    				|| c == 0x0387 // Type 24
    				))
                  ) {
    		      
    		    return false;
            }
    		
    	}
    
    	// All characters passed the tests.
    	return true;
    	
    } // isXMLName(String):boolean
    
    //
    // DEBUG
    //

    /** Prints out the tables. */
    public void print() {

        if (DEBUG_PRINT_TABLES) {
            // This assumes that the document is small
            System.out.println("# start table");
            for (int i = 0; i < fNodeCount; i++) {
                if (i % 10 == 0) {
                    System.out.println("num\ttype\tname\tval\tpar\tfch\tlch\tpsib\tnsib");
                }
                System.out.print(i);
                System.out.print('\t');
                System.out.print(fNodeType[0][i]);
                System.out.print('\t');
                System.out.print(fNodeName[0][i]);
                System.out.print('\t');
                System.out.print(fNodeValue[0][i]);
                System.out.print('\t');
                System.out.print(fNodeParent[0][i]);
                System.out.print('\t');
                System.out.print(fNodeFirstChild[0][i]);
                System.out.print('\t');
                System.out.print(fNodeLastChild[0][i]);
                System.out.print('\t');
                System.out.print(fNodePrevSib[0][i]);
                System.out.print('\t');
                System.out.print(fNodeNextSib[0][i]);
                System.out.println();
            }
            System.out.println("# end table");
        }

    } // print()
    
    //
    // Package methods
    //

    /**
     * Synchronizes the node's children with the internal structure.
     * Fluffing the children at once solves a lot of work to keep
     * the two structures in sync. The problem gets worse when
     * editing the tree -- this makes it a lot easier.
     */
    void synchronizeChildren() {
        
        // perform basic synchronized children
        
        // NOTE: The super method will call synchronizeChildren
        // recursively unless the syncChildren variable is preset
        // to false. -Ac
        
        syncChildren = false;
        super.synchronizeChildren();
        
        // assign doctype and document element
        for (NodeImpl node = firstChild; node != null; node = node.nextSibling) {
            int type = node.getNodeType();
            if (type == Node.DOCUMENT_TYPE_NODE) {
                docType = (DocumentTypeImpl)node;
            }
            else if (type == Node.ELEMENT_NODE) {
                docElement = (ElementImpl)node;

                // NOTE: I would like to break here with the assumption
                // that no document type node would ever appear after a
                // the root element node. However, the DOM specification
                // doesn't happen to mention this so I better not make
                // the assumption. -Ac

                //break;
            }
        }

    } // synchronizeChildren()
    

    //
    // Private methods
    //

    /** Ensures that the internal tables are large enough. */
    private boolean ensureCapacity(int chunk, int index) {

        // create buffers
        if (fNodeType == null) {
            fNodeType       = new byte[INITIAL_CHUNK_COUNT][];
            fNodeName       = new int[INITIAL_CHUNK_COUNT][];
            fNodeValue      = new int[INITIAL_CHUNK_COUNT][];
            fNodeParent     = new int[INITIAL_CHUNK_COUNT][];
            fNodeFirstChild = new int[INITIAL_CHUNK_COUNT][];
            fNodeLastChild  = new int[INITIAL_CHUNK_COUNT][];
            fNodePrevSib    = new int[INITIAL_CHUNK_COUNT][];
            fNodeNextSib    = new int[INITIAL_CHUNK_COUNT][];
        }

        // return true if table is already big enough
        try {
            return fNodeType[chunk][index] != 0;
        } 

        // resize the tables
        catch (ArrayIndexOutOfBoundsException ex) {
            byte[][] newByteArray = new byte[chunk * 2][];
            System.arraycopy(fNodeType, 0, newByteArray, 0, chunk);
            fNodeType = newByteArray;
            int[][] newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeName, 0, newIntArray, 0, chunk);
            fNodeName = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeValue, 0, newIntArray, 0, chunk);
            fNodeValue = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeParent, 0, newIntArray, 0, chunk);
            fNodeParent = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeFirstChild, 0, newIntArray, 0, chunk);
            fNodeFirstChild = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeLastChild, 0, newIntArray, 0, chunk);
            fNodeLastChild = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodePrevSib, 0, newIntArray, 0, chunk);
            fNodePrevSib = newIntArray;
            newIntArray = new int[chunk * 2][];
            System.arraycopy(fNodeNextSib, 0, newIntArray, 0, chunk);
            fNodeNextSib = newIntArray;
        } 
        
        catch (NullPointerException ex) {
            // ignore
        }

        // create chunks
        fNodeType[chunk]       = new byte[CHUNK_SIZE];
        fNodeName[chunk]       = new int[CHUNK_SIZE];
        fNodeValue[chunk]      = new int[CHUNK_SIZE];
        fNodeParent[chunk]     = new int[CHUNK_SIZE];
        fNodeFirstChild[chunk] = new int[CHUNK_SIZE];
        fNodeLastChild[chunk]  = new int[CHUNK_SIZE];
        fNodePrevSib[chunk]    = new int[CHUNK_SIZE];
        fNodeNextSib[chunk]    = new int[CHUNK_SIZE];

        // success
        return true;

    } // ensureCapacity(int,int):boolean

    /** Creates a node of the specified type. */
    private int createNode(short nodeType) {

        // ensure tables are large enough
        int chunk = fNodeCount >> CHUNK_SHIFT;
        int index = fNodeCount & CHUNK_MASK;
        ensureCapacity(chunk, index);

        // initialize node
        fNodeType[chunk][index]       = (byte)nodeType;
        fNodeName[chunk][index]       = -1;
        fNodeValue[chunk][index]      = -1;
        fNodeParent[chunk][index]     = -1;
        fNodeFirstChild[chunk][index] = -1;
        fNodeLastChild[chunk][index]  = -1;
        fNodePrevSib[chunk][index]    = -1;
        fNodeNextSib[chunk][index]    = -1;

        // return node index number
        return fNodeCount++;

    } // createNode(short):int

} // class DocumentImpl
