/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import org.w3c.dom.*;

/**
 * This class represents a Document Type <em>declaraction</em> in
 * the document itself, <em>not</em> a Document Type Definition (DTD).
 * An XML document may (or may not) have such a reference.
 * <P>
 * DocumentType is an Extended DOM feature, used in XML documents but
 * not in HTML.
 * <P>
 * Note that Entities and Notations are no longer children of the
 * DocumentType, but are parentless nodes hung only in their
 * appropriate NamedNodeMaps.
 * <P>
 * This area is UNDERSPECIFIED IN REC-DOM-Level-1-19981001
 * Most notably, absolutely no provision was made for storing
 * and using Element and Attribute information. Nor was the linkage
 * between Entities and Entity References nailed down solidly.
 *
 * @version Revision: 57 1.4 src/com/ibm/xml/dom/DocumentTypeImpl.java, parser, xml4j2, xml4j2_0_6 
 * @since  PR-DOM-Level-1-19980818.
 */
public class DocumentTypeImpl 
    extends NodeImpl 
    implements DocumentType {

    //
    // Data
    //

    /** Entities. */
    private NamedNodeMapImpl entities;
    
    /** Notations. */
    private NamedNodeMapImpl notations;

    // NON-DOM

    /** Elements. */
    private NamedNodeMapImpl elements;

    // lazy-eval

    /** Entities, notations, and elements need to be synchronized. */
    private boolean syncInfo;
    
    //
    // Constructors
    //

    /**
     * This is the Fast constructor. Only the fNodeIndex is given here. 
     * All other data, can be requested from the ownerDocument via the index.
     */
    DocumentTypeImpl(DocumentImpl ownerDocument, int nodeIndex) {
        super(ownerDocument, nodeIndex);

        syncChildren = false;
        syncInfo     = true;

    } // <init>(DocumentImpl,int)

    /** Factory method for creating a document type node. */
    public DocumentTypeImpl(DocumentImpl ownerDocument, String name) {
        super(ownerDocument, name, null);

        // DOM
        entities  = new NamedNodeMapImpl(ownerDocument, null);
        notations = new NamedNodeMapImpl(ownerDocument, null);

        // NON-DOM
        elements = new NamedNodeMapImpl(ownerDocument,null);

    } // <init>(DocumentImpl,String)
  
    //
    // Node methods
    //

    /** 
     * A short integer indicating what type of node this is. The named
     * constants for this value are defined in the org.w3c.dom.Node interface.
     */
    public final short getNodeType() {
        return Node.DOCUMENT_TYPE_NODE;
    }
    
    /**
     * DocumentTypes never have a nodeValue.
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR)
     */
    public void setNodeValue(String value) throws DOMException {
    	throw new DOMExceptionImpl(
    		DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
    }

    /** Clones the node. */
    public Node cloneNode(boolean deep) {

    	DocumentTypeImpl newnode = (DocumentTypeImpl)super.cloneNode(deep);

    	// NamedNodeMaps must be cloned explicitly, to avoid sharing them.
    	newnode.entities  = entities.cloneMap();
    	newnode.notations = notations.cloneMap();
    	newnode.elements  = elements.cloneMap();

    	return newnode;

    } // cloneNode(boolean):Node

    /** Document type is not allowed to have children. */
    public Node appendChild(Node newChild) {
        throw new DOMExceptionImpl(
            DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
    }

    //
    // DocumentType methods
    //

    /**
     * Name of this document type. If we loaded from a DTD, this should
     * be the name immediately following the DOCTYPE keyword.
     */
    public String getName() {

        if (syncData) {
            synchronizeData();
        }
    	return name;

    } // getName():String

    /**
     * Access the collection of general Entities, both external and
     * internal, defined in the DTD. For example, in:
     * <p>
     * <pre>
     *   &lt;!doctype example SYSTEM "ex.dtd" [
     *     &lt;!ENTITY foo "foo"&gt;
     *     &lt;!ENTITY bar "bar"&gt;
     *     &lt;!ENTITY % baz "baz"&gt;
     *     ]&gt;
     * </pre>
     * <p>
     * The Entities map includes foo and bar, but not baz. It is promised that
     * only Nodes which are Entities will exist in this NamedNodeMap.
     * <p>
     * For HTML, this will always be null.
     * <p>
     * Note that "built in" entities such as &amp; and &lt; should be
     * converted to their actual characters before being placed in the DOM's
     * contained text, and should be converted back when the DOM is rendered
     * as XML or HTML, and hence DO NOT appear here.
     */
    public NamedNodeMap getEntities() {
        if (syncInfo) {
            synchronizeInfo();
            }
    	return entities;
    }

    /**
     * Access the collection of Notations defined in the DTD.  A
     * notation declares, by name, the format of an XML unparsed entity
     * or is used to formally declare a Processing Instruction target.
     */
    public NamedNodeMap getNotations() {
        if (syncInfo) {
            synchronizeInfo();
            }
    	return notations;
    }

    //
    // Public methods
    //

    /**
     * NON-DOM: Subclassed to flip the entities' and notations' readonly switch
     * as well.
     * @see NodeImpl.setReadOnly
     */
    public void setReadOnly(boolean readOnly, boolean deep) {
    	
        if (syncInfo) {
            synchronizeInfo();
            }
        setReadOnly(readOnly, deep);

        // set read-only property
        elements.setReadOnly(readOnly, true);
        entities.setReadOnly(readOnly, true);
    	notations.setReadOnly(readOnly, true);

    } // setReadOnly(boolean,boolean)
    
    /**
     * NON-DOM: Access the collection of ElementDefinitions.
     * @see ElementDefinitionImpl
     */
    public NamedNodeMap getElements() {
        if (syncInfo) {
            synchronizeInfo();
            }
    	return elements;
    }

    //
    // Protected methods
    //

    /** Synchronizes the entities, notations, and elements. */
    protected void synchronizeInfo() {

        // create new node maps
        entities  = new NamedNodeMapImpl(ownerDocument, null);
        notations = new NamedNodeMapImpl(ownerDocument, null);
        elements  = new NamedNodeMapImpl(ownerDocument, null);

        // fill node maps
        for (int index = ownerDocument.getFirstChild(fNodeIndex);
            index != -1; 
            index = ownerDocument.getNextSibling(index)) {

            NodeImpl node = ownerDocument.getNodeObject(index);
            int type = node.getNodeType();
            switch (type) {
                
                // internal, external, and unparsed entities
                case Node.ENTITY_NODE: {
                    entities.setNamedItem(node);
                    break;
                }

                // notations
                case Node.NOTATION_NODE: {
                    notations.setNamedItem(node);
                    break;
                }

                // element definitions
                case NodeImpl.ELEMENT_DEFINITION_NODE: {
                    
                    // add element definition
                    elements.setNamedItem(node);

                    // add attributes to element definition
                    NamedNodeMap attrs = node.getAttributes();
                    for (int attrIndex = ownerDocument.getFirstChild(node.fNodeIndex);
                         attrIndex != -1;
                         attrIndex = ownerDocument.getNextSibling(attrIndex)) {
                        NodeImpl attr = ownerDocument.getNodeObject(attrIndex);
                        attrs.setNamedItem(attr);
                    }
                    break;
                }

                // NOTE: Should never get here! -Ac
                default: {
                    System.out.println("DocumentTypeImpl#synchronizeInfo: node.getNodeType() = "+node.getNodeType()+", class = "+node.getClass().getName());
                }
             }
        }

        // no need to synchronize again
        syncInfo = false;

    } // synchronizeInfo()

} // class DocumentTypeImpl
