/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import org.w3c.dom.*;

/**
 * NON-DOM CLASS: Describe one of the Elements (and its associated
 * Attributes) defined in this Document Type.
 * <p>
 * I've included this in Level 1 purely as an anchor point for default
 * attributes. In Level 2 it should enable the ChildRule support.
 *
 * @version Revision: 60 1.2 src/com/ibm/xml/dom/ElementDefinitionImpl.java, parser, xml4j2, xml4j2_0_6 
 */
public class ElementDefinitionImpl 
    extends NodeImpl {

    //
    // Data
    //

    /** Default attributes. */
	private NamedNodeMapImpl attributes;

    // lazy-eval info

    /** Default attribute values need to be synchronized. */
    private boolean syncAttrs;

    //
    // Constructors
    //

    /**
     * This is the Fast constructor. Only the fNodeIndex is given here.
     * All other data, can be requested from the ownerDocument via the index.
     */
    ElementDefinitionImpl(DocumentImpl ownerDocument, int nodeIndex) {
        super(ownerDocument, nodeIndex);
        syncAttrs = true;
    }

    /** Factory constructor. */
    public ElementDefinitionImpl(DocumentImpl ownerDocument, String name) {
    	super(ownerDocument, name, null);
        attributes = new NamedNodeMapImpl(ownerDocument, null);
    }

    //
    // Node methods
    //

    /** 
     * A short integer indicating what type of node this is. The named
     * constants for this value are defined in the org.w3c.dom.Node interface.
     */
    public final short getNodeType() {
        return NodeImpl.ELEMENT_DEFINITION_NODE;
    }

    /**
     * Replicate this object.
     */
    public Node cloneNode(boolean deep) {

    	ElementDefinitionImpl newnode = (ElementDefinitionImpl) super.cloneNode(deep);
    	// NamedNodeMap must be explicitly replicated to avoid sharing
    	newnode.attributes = attributes.cloneMap();
    	return newnode;

    } // cloneNode(boolean):Node

    /**
     * Query the attributes defined on this Element.
     * <p>
     * In the base implementation this Map simply contains Attribute objects
     * representing the defaults. In a more serious implementation, it would
     * contain AttributeDefinitionImpl objects for all declared Attributes,
     * indicating which are Default, DefaultFixed, Implicit and/or Required.
     * 
     * @return org.w3c.dom.NamedNodeMap containing org.w3c.dom.Attribute
     */
    public NamedNodeMap getAttributes() {

        if (syncAttrs) {
            synchronizeAttributes();
        }
    	return attributes;

    } // getAttributes():NamedNodeMap

    //
    // Private methods
    //

    /** Synchronizes the default attribute values. */
    private void synchronizeAttributes() {

        // REVISIT: Make NamedNodeMap implementation "fast". -Ac

        // create attributes node map
        attributes = new NamedNodeMapImpl(ownerDocument, null);

        // Default attributes dangle as children of the element
        // definition "node" in the internal fast table.
        for (int nodeIndex = ownerDocument.getFirstChild(fNodeIndex);
             nodeIndex != -1;
             nodeIndex = ownerDocument.getNextSibling(nodeIndex)) {
            Node attr = ownerDocument.getNodeObject(nodeIndex);
            attributes.setNamedItem(attr);
        }

        // attributes are now synced
        syncAttrs = false;

    } // synchronizeAttributes()

} // class ElementDefinitionImpl
