/*
 * (C) Copyright IBM Corp. 1998,1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dom;

import org.w3c.dom.*;

/**
 * Text nodes hold the non-markup, non-Entity content of
 * an Element or Attribute.
 * <P>
 * When a document is first made available to the DOM, there is only
 * one Text object for each block of adjacent plain-text. Users (ie,
 * applications) may create multiple adjacent Texts during editing --
 * see {@link Element#normalize} for discussion.
 * <P>
 * Note that CDATASection is a subclass of Text. This is conceptually
 * valid, since they're really just two different ways of quoting
 * characters when they're written out as part of an XML stream.
 *
 * @version Revision: 69 1.4 src/com/ibm/xml/dom/TextImpl.java, parser, xml4j2, xml4j2_0_6 
 * @since  PR-DOM-Level-1-19980818.
 */
public class TextImpl 
    extends CharacterDataImpl 
    implements Text {

    //
    // Constructors
    //

    /**
     * This is the Fast constructor. Only the fNodeIndex is given here.
     * All other data, can be requested from the ownerDocument via the index.
     */
    TextImpl(DocumentImpl ownerDocument, int nodeIndex) {
        super(ownerDocument, nodeIndex);
    }
  
    /** Factory constructor. */
    public TextImpl(DocumentImpl ownerDoc, String data) {
        super(ownerDoc, data);
    }  
    
    //
    // Node methods
    //

    /** 
     * A short integer indicating what type of node this is. The named
     * constants for this value are defined in the org.w3c.dom.Node interface.
     */
    public short getNodeType() {
        return Node.TEXT_NODE;
    }

    /** Returns the node name. */
    public String getNodeName() {
        return "#text";
    }
    
    boolean isIgnorableWhitespace = false;
    /**
     * NON-DOM: Returns whether this Text is ignorable whitespace.
     */
    public boolean isIgnorableWhitespace() {
        if (syncChildren) {
            synchronizeChildren();
        }
        return this.isIgnorableWhitespace;
    }
    
    /**
     * NON-DOM: Set whether this Text is ignorable whitespace.
     */
    public void setIgnorableWhitespace(boolean isIgnorableWhitespace) {
        if (syncChildren) {
            synchronizeChildren();
        }
        this.isIgnorableWhitespace = isIgnorableWhitespace;
    }
    

    /** Clones this node. */
    public Node cloneNode(boolean deep) {
        return ownerDocument.createTextNode(getNodeValue());
    }  

    //
    // Text methods
    //

    /** 
     * Break a text node into two sibling nodes.  (Note that if the
     * current node has no parent, they won't wind up as "siblings" --
     * they'll both be orphans.)
     *
     * @param offset The offset at which to split. If offset is at the
     * end of the available data, the second node will be empty.
     *
     * @returns A reference to the new node (containing data after the
     * offset point). The original node will contain data up to that
     * point.
     *
     * @throws DOMException(INDEX_SIZE_ERR) if offset is <0 or >length.
     *
     * @throws DOMException(NO_MODIFICATION_ALLOWED_ERR) if node is read-only.
     */
    public Text splitText(int offset) 
        throws DOMException {

    	if (readOnly) {
    		throw new DOMExceptionImpl(
    			DOMException.NO_MODIFICATION_ALLOWED_ERR, null);
        }

        if (syncData) {
            synchronizeData();
        }
    	if (offset < 0 || offset > value.length() - 1) {
    		throw new DOMExceptionImpl(DOMException.INDEX_SIZE_ERR, null);
        }
    		
        // split text into two separate nodes
    	Text newText = ownerDocument.createTextNode(value.substring(offset));
    	setNodeValue(value.substring(0, offset));

        // insert new text node
    	if (parentNode != null) {
    		parentNode.insertBefore(newText, nextSibling);
        }

    	return newText;

    } // splitText(int):Text

    //
    // Private methods
    //

    /** Synchronizes the underlying data. */
    protected void synchronizeData() {

        // get initial text value
        value = ownerDocument.getNodeValue(fNodeIndex);
        int realNext = ownerDocument.getRealNextSibling(fNodeIndex);
        int parent   = ownerDocument.getParentNode(fNodeIndex);

        // revisit: we only normalize text nodes with Elements as parents.
        if (parent != -1 && ownerDocument.getNodeType(parent) == Node.ELEMENT_NODE) {
            int type = ownerDocument.getNodeType(realNext);
            if (realNext != -1 && type == Node.TEXT_NODE) {
                StringBuffer sb = new StringBuffer(value);
                while (realNext != -1 && type == Node.TEXT_NODE) {
                    sb.append(ownerDocument.getNodeValue(realNext));
                    realNext = ownerDocument.getRealNextSibling(realNext);
                    type = ownerDocument.getNodeType(realNext);
                }
                value = sb.toString();
            }
        }

        // no need for future synchronizations
        syncData = false;

    } // synchronizeData()
    
    /**
     * The TextImpl doesn't have any children and we are using the firstchild
     * to hold a 1 or 0, to tell whether it's ignorable whitespace.
     */
    void synchronizeChildren() {
        isIgnorableWhitespace = ownerDocument.getFirstChild(fNodeIndex) == 1;
        syncChildren = false;                        
    }

} // class TextImpl
